#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include "bandwidthd.h"


// We must call regular exit to write out profile data, but child forks are supposed to usually
// call _exit?
#ifdef PROFILE
#define _exit(x) exit(x)
#endif

/*
#ifdef DEBUG
#define fork() (0)
#endif
*/

// ****************************************************************************************
// ** Global Variables
// ****************************************************************************************

static pcap_t *pd;

unsigned int GraphIntervalCount = 0;
unsigned int IpCount = 0;
unsigned int SubnetCount = 0;
time_t IntervalStart;
    
struct SubnetData SubnetTable[SUBNET_NUM];
struct IPData IpTable[IP_NUM];

key_t IPCKey;
int shmid;
struct IPCData *IPCSharedData;

struct IPDataStore *IPDataStore = NULL;
extern int yyparse(void);
extern FILE *yyin;

struct config config;

int main(int argc, char **argv)
    {
    struct bpf_program fcode;
    bpf_u_int32 localnet, netmask;    
    //RETSIGTYPE (*oldhandler)(int);
    u_char *pcap_userdata = 0;
    struct shmid_ds shmstatus;
	char Error[PCAP_ERRBUF_SIZE];

	config.dev = NULL;
	config.skip_intervals = CONFIG_GRAPHINTERVALS;
	config.graph_cutoff = CONFIG_GRAPHCUTOFF;
	config.promisc = TRUE;
	config.graph = TRUE;
	config.output_cdf = FALSE;
	yyin = fopen("./etc/bandwidthd.conf", "r");
	yyparse();

	if (fork2())
		exit(0);

	IPCKey = ftok(".", 'a');
	if ((shmid = shmget(IPCKey, sizeof(struct IPCData)*IP_NUM, IPC_CREAT | IPC_EXCL)) == -1)
		{
		if ((shmid = shmget(IPCKey, sizeof(struct IPCData)*IP_NUM, 0)) == -1)
			{
			printf("Error allocating %d bytes of IPC Shared memory. Do you have System V IPC turned on in your kernel? And do you have enough?\n", sizeof(struct IPCData)*IP_NUM);
			exit(0);
			}

		shmctl(shmid, IPC_STAT, &shmstatus);
		if (shmstatus.shm_nattch == 0)
			{
			shmctl(shmid, IPC_RMID, &shmstatus);
			if ((shmid = shmget(IPCKey, sizeof(struct IPCData)*IP_NUM, IPC_CREAT | IPC_EXCL)) == -1)
				{
				printf("Shared memory busted.  I am busted, everything is busted.  Have a nice day\n");
				exit(1);
				}
			}
		else
			{
			printf("My shared memory segment %d is already in use (%ld locks), perhaps bandwidthd is already running in this directory?\n", shmid, shmstatus.shm_nattch);
			exit(1);
			}
		}
	
	if ((int) (IPCSharedData = shmat(shmid, 0, 0)) == -1)
		{
		printf("Error attaching to shared memory segment!\n");
		exit(0);
		}

    IntervalStart = time(NULL);
	
	pd = pcap_open_live(config.dev, 100, config.promisc, 1000, Error);
        if (pd == NULL) printf("%s\n", Error);
                                                           
    if (pcap_lookupnet(config.dev, &localnet, &netmask, NULL) < 0) {
        localnet = 0;
        netmask = 0;
		printf("%s\n", Error);
        pcap_perror(pd, "Error");
        }
              
    if (pcap_compile(pd, &fcode, "ip", 1, netmask) < 0)
        pcap_perror(pd, "Error");

    if (pcap_setfilter(pd, &fcode) < 0)
        pcap_perror(pd, "Error");
                                           
    if (pcap_loop(pd, -1, PacketCallback, pcap_userdata) < 0) {
        (void)fprintf(stderr, "Bandwidthd: pcap_loop: %s\n",  pcap_geterr(pd));
        exit(1);
        }

    pcap_close(pd);
    exit(0);        
    }
   
void PacketCallback(u_char *user, const struct pcap_pkthdr *h, const u_char *p)
    {
    unsigned int counter;

    u_int caplen = h->caplen;
    const struct ip *ip;

    uint32_t srcip;
    uint32_t dstip;


    struct IPData *ptrIPData;

    if (h->ts.tv_sec > IntervalStart + INTERVAL)  // Then write out this intervals data and possibly kick off the grapher
        {
        GraphIntervalCount++;
        CommitData(IntervalStart+INTERVAL);
		IpCount = 0;
        IntervalStart=h->ts.tv_sec;
        }

    caplen -= sizeof(struct ether_header);  // We're only measuring ip size, so pull off the ethernet header
    p += sizeof(struct ether_header); // Move the pointer past the ethernet header

    ip = (const struct ip *)p; // Point ip at the ip header

    srcip = ntohl(*(uint32_t *) (&ip->ip_src));
    dstip = ntohl(*(uint32_t *) (&ip->ip_dst));

    for (counter = 0; counter < SubnetCount; counter++)
        {
		// Packets from a monitored subnet to a monitored subnet will be
		// credited to both ip's

        if (SubnetTable[counter].ip == (srcip & SubnetTable[counter].mask))
            {
            ptrIPData = FindIp(srcip);  // Return or create this ip's data structure
			if (ptrIPData)
	            Credit(&(ptrIPData->Send), ip);
            }
    
        if (SubnetTable[counter].ip == (dstip & SubnetTable[counter].mask))
            {
            ptrIPData = FindIp(dstip);
    		if (ptrIPData)
		        Credit(&(ptrIPData->Receive), ip);
            }                        
        }
    }

inline void Credit(struct Statistics *Stats, const struct ip *ip)
    {
    unsigned long size;
    const struct tcphdr *tcp;
    uint16_t sport, dport;

    size = ntohs(ip->ip_len);

    Stats->total += size;
    
    switch(ip->ip_p)
        {
        case 6:     // TCP
            tcp = (struct tcphdr *)(ip+1);
			tcp = (struct tcphdr *) ( ((char *)tcp) + ((ip->ip_hl-5)*4) ); // Compensate for IP Options
            Stats->tcp += size;
#ifdef SOLARIS
            sport = ntohs(tcp->th_sport);
            dport = ntohs(tcp->th_dport);			
#else
            sport = ntohs(tcp->source);
            dport = ntohs(tcp->dest);
#endif
            if (sport == 80 || dport == 80)
                Stats->http += size;

            if (sport == 1044|| dport == 1044||		// Direct File Express
				sport == 1045|| dport == 1045|| 	// ''  <- Dito Marks
                sport == 1214|| dport == 1214||		// Grokster, Kaza, Morpheus
				sport == 4661|| dport == 4661||		// EDonkey 2000
				sport == 4662|| dport == 4662||     // ''
				sport == 4665|| dport == 4665||     // ''
				sport == 5190|| dport == 5190||		// Song Spy
				sport == 5500|| dport == 5500||		// Hotline Connect
				sport == 5501|| dport == 5501||		// ''
				sport == 5502|| dport == 5502||		// ''
				sport == 5503|| dport == 5503||		// ''
				sport == 6346|| dport == 6346||		// Gnutella Engine
				sport == 6347|| dport == 6347||		// ''
				sport == 6666|| dport == 6666||		// Yoink
				sport == 6667|| dport == 6667||		// ''
				sport == 7788|| dport == 7788||		// Budy Share
				sport == 8888|| dport == 8888||		// AudioGnome, OpenNap, Swaptor
				sport == 8889|| dport == 8889||		// AudioGnome, OpenNap
				sport == 28864|| dport == 28864||	// hotComm				
				sport == 28865|| dport == 28865)	// hotComm
                Stats->p2p += size;
            break;
        case 17:
            Stats->udp += size;
            break;
        case 1: 
            Stats->icmp += size;
            break;
        }
    }

// TODO:  Throw away old data!
void DropOldData(long int timestamp) 	// Go through the ram datastore and dump old data
	{
	struct IPDataStore *DataStore;
	struct IPDataStore *PrevDataStore;	
	struct DataStoreBlock *DeletedBlock;
	
	PrevDataStore = NULL;
    DataStore = IPDataStore;

	while(DataStore)  // we have data
		{
        while(DataStore->FirstBlock->LatestTimestamp < timestamp - RANGE1)
			{
            if ((!DataStore->FirstBlock->Next) && PrevDataStore) // There is no valid block of data for this ip, so unlink the whole ip
				{ 												// Don't bother unlinking the ip if it's the first one, that's to much
																// Trouble
				PrevDataStore->Next = DataStore->Next;	// Unlink the node
				free(DataStore->FirstBlock->Data);      // Free the memory
				free(DataStore->FirstBlock);
				free(DataStore);												
				DataStore = PrevDataStore;				// Rewind back to the prev data store so the move next below doesn't skip a node		
				}				
			else // Just unlink this block
				{
				DeletedBlock = DataStore->FirstBlock;
				DataStore->FirstBlock = DataStore->FirstBlock->Next;	// Unlink the block
				free(DeletedBlock->Data);
				free(DeletedBlock);
			    }
			}

		PrevDataStore = DataStore;				
		DataStore = DataStore->Next;
		}
	}

void StoreIPDataInRam(struct IPData *IPData)
	{
	struct IPDataStore *DataStore;
	struct DataStoreBlock *DataStoreBlock;

	if (!IPDataStore) // we need to create the first entry
		{
		// Allocate Datastore for this IP
        IPDataStore = malloc(sizeof(struct IPDataStore));
				
		IPDataStore->ip = IPData->ip;
		IPDataStore->Next = NULL;
					
		// Allocate it's first block of storage
		IPDataStore->FirstBlock = malloc(sizeof(struct DataStoreBlock));
		IPDataStore->FirstBlock->LatestTimestamp = 0;
		IPDataStore->FirstBlock->NumEntries = 0;
		IPDataStore->FirstBlock->Data = calloc(IPDATAALLOCCHUNKS, sizeof(struct IPData));
		IPDataStore->FirstBlock->Next = NULL;																		
		}

	DataStore = IPDataStore;

	// Take care of first case
	
	while (DataStore) // Is not null
		{
		if (DataStore->ip == IPData->ip) // then we have the right store
			{
			DataStoreBlock = DataStore->FirstBlock;

			while(DataStoreBlock) // is not null
				{
				if (DataStoreBlock->NumEntries < IPDATAALLOCCHUNKS) // We have a free spot
					{
					memcpy(&DataStoreBlock->Data[DataStoreBlock->NumEntries++], IPData, sizeof(struct IPData));
					DataStoreBlock->LatestTimestamp = IPData->timestamp;
					return;
					}
                else
					{
					if (!DataStoreBlock->Next) // there isn't another block, add one
						{
                        DataStoreBlock->Next = malloc(sizeof(struct DataStoreBlock));
						DataStoreBlock->Next->LatestTimestamp = 0;
						DataStoreBlock->Next->NumEntries = 0;
						DataStoreBlock->Next->Data = calloc(IPDATAALLOCCHUNKS, sizeof(struct IPData));
						DataStoreBlock->Next->Next = NULL;																				
						}

					DataStoreBlock = DataStoreBlock->Next;
					}
				}						

			return;
			}
		else
			{
			if (!DataStore->Next) // there is no entry for this ip, so lets make one.
				{
				// Allocate Datastore for this IP
                DataStore->Next = malloc(sizeof(struct IPDataStore));
				
		        DataStore->Next->ip = IPData->ip;
				DataStore->Next->Next = NULL;
					
				// Allocate it's first block of storage
				DataStore->Next->FirstBlock = malloc(sizeof(struct DataStoreBlock));
				DataStore->Next->FirstBlock->LatestTimestamp = 0;
				DataStore->Next->FirstBlock->NumEntries = 0;
				DataStore->Next->FirstBlock->Data = calloc(IPDATAALLOCCHUNKS, sizeof(struct IPData));
				DataStore->Next->FirstBlock->Next = NULL;																		
				}

			DataStore = DataStore->Next;			
			}
		}
	}

void CommitData(long int timestamp)
    {
    unsigned int counter;
	int NumGraphs = 0;

	struct IPDataStore *DataStore;

	for (counter=0; counter < IpCount; counter++)
        {
        IpTable[counter].timestamp = timestamp;
		StoreIPDataInRam(&IpTable[counter]);
        }

	if (GraphIntervalCount%config.skip_intervals == 0)
		{
		// Reap a couple zombies
		waitpid(-1, NULL, WNOHANG);
		waitpid(-1, NULL, WNOHANG);

		// break off from the main line so we don't miss any packets while we graph
		DataStore = IPDataStore;
		if (DataStore)
			{
			if (!fork()) // if there is a datastore to graph fork, and if we're the child, graph it.
				{
#ifdef PROFILE
				// Got this incantation from a message board.  Don't forget to set
				// GMON_OUT_PREFIX in the shell
				extern void _start(void), etext(void);
				printf("Calling profiler startup...\n");
				monstartup((u_long) &_start, (u_long) &etext);
#endif
     	        nice(4); // reduce priority so I don't choke out other tasks
			
				DataStore = IPDataStore;	
				while (DataStore) // Is not null
					{
					if (DataStore->FirstBlock->NumEntries > 0)
						GraphIp(DataStore, &IPCSharedData[NumGraphs++], timestamp+LEAD*RANGE1);
			        DataStore = DataStore->Next;
					}

				MakeIndexPages(NumGraphs);
	
				_exit(0);
				}
			else	// drop the old data in the main thread
				DropOldData(timestamp);
			}	
		}
    }

// ****** FindIp **********
// ****** Returns or allocates an Ip's data structure

inline struct IPData *FindIp(uint32_t ipaddr)
    {
    unsigned int counter;
    
    for (counter=0; counter < IpCount; counter++)
        if (IpTable[counter].ip == ipaddr)
            return (&IpTable[counter]);
    
    if (IpCount >= IP_NUM)
        {
        printf("IP_NUM is too low, dropping ip....");
       	return(NULL);
        }
	
    memset(&IpTable[IpCount], 0, sizeof(struct IPData));
    IpTable[IpCount].Magick[0] = 'R';
    IpTable[IpCount].Magick[1] = 'E';
    IpTable[IpCount].Magick[2] = 'C';
    IpTable[IpCount].Magick[3] = 'B';

    IpTable[IpCount].ip = ipaddr;
    return (&IpTable[IpCount++]);
    }

size_t ICGrandTotalDataPoints = 0;

char inline *HostIp2CharIp(unsigned long ipaddr, char *buffer)
    {
	struct in_addr in_addr;
	char *s;

	in_addr.s_addr = htonl(ipaddr);	
    s = inet_ntoa(in_addr);
	strncpy(buffer, s, 16);
	buffer[15] = '\0';
	return(buffer);
/*  uint32_t ip = *(uint32_t *)ipaddr;

	sprintf(buffer, "%d.%d.%d.%d", (ip << 24)  >> 24, (ip << 16) >> 24, (ip << 8) >> 24, (ip << 0) >> 24);
*/
    }

// Add better error checking

int fork2()
    {
    pid_t pid;

    if (!(pid = fork()))
        {
        if (!fork())
        	{
#ifdef PROFILE
				// Got this incantation from a message board.  Don't forget to set
				// GMON_OUT_PREFIX in the shell
				extern void _start(void), etext(void);
				printf("Calling profiler startup...\n");
				monstartup((u_long) &_start, (u_long) &etext);
#endif
            return(0);
            }        

        _exit(0);
        }
    
    waitpid(pid, NULL, 0);
    return(1);
    }

