#include <unistd.h>
#include <setjmp.h>
#include <signal.h>
#include <string.h>
#include <netdb.h>
#include <sys/socket.h>
#include <resolv.h>
#include "bandwidthd.h"

extern struct IPCData *IPCSharedData;
extern unsigned int SubnetCount;
extern struct config config;

char YLegend;
long int Divisor;
jmp_buf dnsjump;

static void rdnslngjmp(int signal);

void rdns(char *Buffer, unsigned long IP)  // This takes over sigalarm!
	{
	char DNSError[] = "DNS Timeout: Correct to speed up graphing";
	char None[] = "Configure DNS to reverse this IP";
	char TooManyDNSTimeouts[] = "Too many dns timeouts, reverse lookups suspended";
	struct hostent *hostent;
	char chrIP[50];
	static int Init = TRUE;
	static int DNSTimeouts = 0;  // This is reset for each run because we're forked
	unsigned long addr = htonl(IP);

    _res.retrans = 1;
    _res.retry = 2;

	if (Init)
		{
        signal(SIGALRM, rdnslngjmp);
		Init = FALSE;
		}

	if (DNSTimeouts > 100)
		{
		printf("Too many dns timeouts, reverse lookups suspended\n");
        strncpy(Buffer, TooManyDNSTimeouts, 253);
		Buffer[254] = '\0';
		return;
		}		

	if (setjmp(dnsjump) == 0)
		{
		alarm(10);  // Don't let gethostbyaddr hold us up too long
		hostent = gethostbyaddr((char *) &addr, 4, AF_INET); // (char *)&Data->IP				
		alarm(0);
		
		if (hostent)
			sprintf(Buffer, "%s", hostent->h_name);
		else
			{
	        strncpy(Buffer, None, 253);
			Buffer[254] = '\0';
			}
		}
	else  // Our alarm timed out
		{
		HostIp2CharIp(IP, chrIP);
		printf("DNS timeout for %s: This problem reduces bandwidthd performance\n", chrIP);
		DNSTimeouts++;
        strncpy(Buffer, DNSError, 253);
		Buffer[254] = '\0';
		}
	}

static void rdnslngjmp(int signal)
	{
    longjmp(dnsjump, 1);
	}

// QUick sort Descending
void QuickSortIpcData(struct IPCData *IPCData[], int low, int high)
	{
	struct IPCData *Tmp;
    int partition, bottom, top;

	if (low >= high) return;

	if ((high - low) == 1)
		{
        if (IPCData[low]->Total < IPCData[high]->Total)
			{
			// Flip 'em
			Tmp = IPCData[low];
			IPCData[low] = IPCData[high];
			IPCData[high] = Tmp; 	
			}
		}

    // Pick a partition element and move it to the end of the array
	partition = (high+low)/2;
    Tmp = IPCData[high];
	IPCData[high] = IPCData[partition];
	IPCData[partition] = Tmp;

   	bottom = low;
	top = high-1;

	while(bottom < top)
		{
		//                                              Partition Element
		while(bottom < top && IPCData[bottom]->Total >= IPCData[high]->Total)					
			bottom++;

		while(bottom < top && IPCData[top]->Total <= IPCData[high]->Total)					
			top--;

		if (bottom < top)
				{
				// Flip 'em
				Tmp = IPCData[bottom];
				IPCData[bottom] = IPCData[top];
				IPCData[top] = Tmp; 	
				}
		}
	
	// Put the pivot in it's final spot
    Tmp = IPCData[high];
	IPCData[high] = IPCData[bottom];
	IPCData[bottom] = Tmp;

	QuickSortIpcData(IPCData, low, bottom-1);
	QuickSortIpcData(IPCData, bottom+1, high);	
	}

void OutputCdfLine(FILE *cdf, struct IPCData *Data)
	{
	char IPBuffer[50];

	if (Data->Total > config.graph_cutoff) // Qualifies for graphing
		{
		HostIp2CharIp(Data->IP, IPBuffer);
		fprintf(cdf, "%s,%lu,%llu,%llu,%llu,%llu,%llu,%llu,%llu,%llu\n", IPBuffer, time(NULL), Data->Total, Data->TotalSent, Data->TotalReceived, Data->HTTP, Data->P2P, Data->TCP, Data->UDP, Data->ICMP);
		}
	}

void PrintTableLine(FILE *stream, struct IPCData *Data)
	{
	char Buffer1[50];
	char Buffer2[50];
	char Buffer3[50];
	char Buffer4[50];
	char Buffer5[50];
	char Buffer5b[50];
	char Buffer6[50];
	char Buffer7[50];
	char Buffer8[50];

	// First convert the info to nice, human readable stuff
	HostIp2CharIp(Data->IP, Buffer1);

	if ((Data->Total) < 1024)
		snprintf(Buffer2, 50, "%.1f Bytes", (float)Data->Total);
	else if ((Data->Total) < 1024*1024)
		snprintf(Buffer2, 50, "%.1f KBytes", (float)Data->Total/1024.0);
	else snprintf(Buffer2, 50, "%.1f MBytes", (float)Data->Total/(1024.0*1024.0));

	if ((Data->TotalSent) < 1024)
		snprintf(Buffer3, 50, "%.1f Bytes", (float)Data->TotalSent);
	else if ((Data->TotalSent) < 1024*1024)
		snprintf(Buffer3, 50, "%.1f KBytes", (float)Data->TotalSent/1024.0);
	else snprintf(Buffer3, 50, "%.1f MBytes", (float)Data->TotalSent/(1024.0*1024.0));

	if ((Data->TotalReceived) < 1024)
		snprintf(Buffer4, 50, "%.1f Bytes", (float)Data->TotalReceived);
	else if ((Data->TotalReceived) < 1024*1024)
		snprintf(Buffer4, 50, "%.1f KBytes", (float)Data->TotalReceived/1024.0);
	else snprintf(Buffer4, 50, "%.1f MBytes", (float)Data->TotalReceived/(1024.0*1024.0));

	if ((Data->HTTP) < 1024)
		snprintf(Buffer5, 50, "%.1f Bytes", (float)Data->HTTP);
	else if ((Data->HTTP) < 1024*1024)
		snprintf(Buffer5, 50, "%.1f KBytes", (float)Data->HTTP/1024.0);
	else snprintf(Buffer5, 50, "%.1f MBytes", (float)Data->HTTP/(1024.0*1024.0));
	
	if ((Data->P2P) < 1024)
		snprintf(Buffer5b, 50, "%.1f Bytes", (float)Data->P2P);
	else if ((Data->P2P) < 1024*1024)
		snprintf(Buffer5b, 50, "%.1f KBytes", (float)Data->P2P/1024.0);
	else snprintf(Buffer5b, 50, "%.1f MBytes", (float)Data->P2P/(1024.0*1024.0));

	if ((Data->TCP) < 1024)
		snprintf(Buffer6, 50, "%.1f Bytes", (float)Data->TCP);
	else if ((Data->TCP) < 1024*1024)
		snprintf(Buffer6, 50, "%.1f KBytes", (float)Data->TCP/1024.0);
	else snprintf(Buffer6, 50, "%.1f MBytes", (float)Data->TCP/(1024.0*1024.0));

	if ((Data->UDP) < 1024)
		snprintf(Buffer7, 50, "%.1f Bytes", (float)Data->UDP);
	else if ((Data->UDP) < 1024*1024)
		snprintf(Buffer7, 50, "%.1f KBytes", (float)Data->UDP/1024.0);
	else snprintf(Buffer7, 50, "%.1f MBytes", (float)Data->UDP/(1024.0*1024.0));

	if ((Data->ICMP) < 1024)
		snprintf(Buffer8, 50, "%.1f Bytes", (float)Data->ICMP);
	else if ((Data->ICMP) < 1024*1024)
		snprintf(Buffer8, 50, "%.1f KBytes", (float)Data->ICMP/1024.0);
	else snprintf(Buffer8, 50, "%.1f MBytes", (float)Data->ICMP/(1024.0*1024.0));

	if (Data->Graph)
		fprintf(stream, "<TR><TD><a href=\"#%s\">%s</a><TD>%s<TD>%s<TD>%s<TD>%s<TD>%s<TD>%s<TD>%s<TD>%s\n",
			Buffer1, // Ip
			Buffer1, // Ip
			Buffer2, // Total
			Buffer3, // TotalSent
			Buffer4, // TotalReceived
			Buffer5, // HTTP
			Buffer5b, // P2P
			Buffer6, // TCP
			Buffer7, // UDP
			Buffer8); // ICMP
	else
		fprintf(stream, "<TR><TD>%s<TD>%s<TD>%s<TD>%s<TD>%s<TD>%s<TD>%s<TD>%s<TD>%s\n",
			Buffer1, // Ip
			Buffer2, // Total
			Buffer3, // TotalSent
			Buffer4, // TotalReceived
			Buffer5, // HTTP
			Buffer5b, // P2P		
			Buffer6, // TCP
			Buffer7, // UDP
			Buffer8); // ICMP
	}

void MakeIndexPages(int NumIps)
	{
	int SubnetCounter;
	int Counter;
		
	FILE *file, *cdf;
	struct IPCData **IPCData;

	char Buffer1[50];
	char Buffer2[50];
	char HostName[255];
	
	////////////////////////////////////////////////
	// Allocate Index space and sort IP's by traffic

	IPCData = malloc(sizeof(struct IPCData)*IP_NUM);
	// Initilize list
	for (Counter = 0; Counter < NumIps; Counter++)
		{
        IPCData[Counter] = &IPCSharedData[Counter];
		}

	QuickSortIpcData(IPCData, 0, NumIps-1);

	////////////////////////////////////////////////
	// Print main index page
	
	if ((file = fopen("./htdocs/index.html", "w")) == NULL)
		{
		printf("Failed to open ./htdocs/index.html\n");
		exit(1);
		}

	fprintf(file, "<HTML><HEAD>\n<META HTTP-EQUIV=\"REFRESH\" content=\"150\">\n<META HTTP-EQUIV=\"EXPIRES\" content=\"-1\">\n");
	fprintf(file, "<META HTTP-EQUIV=\"PRAGMA\" content=\"no-cache\">\n");
	fprintf(file, "</HEAD><BODY><center><img src=\"logo.gif\"><BR><BR>\nPick a Subnet:<BR>\n");
	
	fprintf(file, "- <a href=\"index.html\">Top20</a> -");
	for (Counter = 0; Counter < SubnetCount; Counter++)            
		{
		HostIp2CharIp(SubnetTable[Counter].ip, Buffer1);
		fprintf(file, "- <a href=Subnet-%s.html>%s</a> -", Buffer1, Buffer1);
		}

	/////  TOP 20

	fprintf(file, "<H1>Top 20 IPs by Traffic</H1></center>Programmed by David Hinkle, Payed for by <a href=http://www.derbytech.com>DerbyTech</a> wireless networking.<BR><center>\n<table width=100%% border=1 cellspacing=0>\n");

    // PASS 1:  Write out the table

	fprintf(file, "<TR><TD>Ip and Name<TD>Total<TD>Total Sent<TD>Total Received<TD>HTTP<TD>P2P<TD>TCP<TD>UDP<TD>ICMP\n");
	for (Counter=0; Counter < 20 && Counter < NumIps; Counter++)
		PrintTableLine(file, IPCData[Counter]);

	fprintf(file, "</table></center>\n");

	// PASS 2: The graphs
	for (Counter=0; Counter < 20 && Counter < NumIps; Counter++)
		if (IPCData[Counter]->Graph)
			{
			HostIp2CharIp(IPCData[Counter]->IP, Buffer1);
			rdns(HostName, IPCData[Counter]->IP);
			fprintf(file, "<a name=\"%s\"><H1><a href=\"#top\">(Top)</a> %s - %s</H1><BR>\nSend:<br>\n<img src=%s-S.png><BR>\n<img src=legend.gif><br>\nReceived:<br>\n<img src=%s-R.png><BR>\n<img src=legend.gif><br>\n<BR>\n", Buffer1, Buffer1, HostName, Buffer1, Buffer1);					
			}

	fprintf(file, "</BODY></HTML>\n");

	fclose(file);

	/////////////////////////////////////////////////
	// Handle Logging

	if (config.output_cdf)
		{
		cdf = fopen("log.cdf", "a");

		for (Counter=0; Counter < NumIps; Counter++)
			{
			OutputCdfLine(cdf, IPCData[Counter]);
			}
		fclose(cdf);
		}

	////////////////////////////////////////////////
	// Print each subnet page

	for (SubnetCounter = 0; SubnetCounter < SubnetCount; SubnetCounter++)
		{
		HostIp2CharIp(SubnetTable[SubnetCounter].ip, Buffer1);
		sprintf(Buffer2, "./htdocs/Subnet-%s.html", Buffer1);
		file = fopen(Buffer2, "w");
		fprintf(file, "<HTML><HEAD>\n<META HTTP-EQUIV=\"REFRESH\" content=\"150\">\n<META HTTP-EQUIV=\"EXPIRES\" content=\"-1\">\n");
		fprintf(file, "<META HTTP-EQUIV=\"PRAGMA\" content=\"no-cache\">\n</HEAD>\n<BODY><CENTER><a name=\"Top\">");
		fprintf(file, "<img src=\"logo.gif\"><BR><BR>\nPick a Subnet:<BR>\n");

		fprintf(file, "- <a href=\"index.html\">Top20</a> -");
		for (Counter = 0; Counter < SubnetCount; Counter++)
			{
			HostIp2CharIp(SubnetTable[Counter].ip, Buffer2);
			fprintf(file, "- <a href=Subnet-%s.html>%s</a> -", Buffer2, Buffer2);
			}

		fprintf(file, "<H1>%s</H1></center>Programmed by David Hinkle, Payed for by <a href=http://www.derbytech.com>DerbyTech</a> wireless networking.<BR></center>\n<table width=100%% border=1 cellspacing=0>\n", Buffer1);


        // PASS 1:  Write out the table

		fprintf(file, "<TR><TD>Ip and Name<TD>Total<TD>Total Sent<TD>Total Received<TD>HTTP<TD>P2P<TD>TCP<TD>UDP<TD>ICMP\n");
		for (Counter=0; Counter < NumIps; Counter++)
			{
            if (SubnetTable[SubnetCounter].ip == (IPCData[Counter]->IP & SubnetTable[SubnetCounter].mask))
				{ // The ip belongs to this subnet
				PrintTableLine(file, IPCData[Counter]);
    			}
			}

		fprintf(file, "</table></center>\n");

		// PASS 2: The graphs
		for (Counter=0; Counter < NumIps && config.graph; Counter++)
			{
            if (SubnetTable[SubnetCounter].ip == (IPCData[Counter]->IP & SubnetTable[SubnetCounter].mask))
				{ // The ip belongs to this subnet
				if (IPCData[Counter]->Graph)
					{
					HostIp2CharIp(IPCData[Counter]->IP, Buffer1);
					rdns(HostName, IPCData[Counter]->IP);
					fprintf(file, "<a name=\"%s\"><H1><a href=\"#top\">(Top)</a> %s - %s</H1><BR>\nSend:<br>\n<img src=%s-S.png><BR>\n<img src=legend.gif><br>\nReceived:<br>\n<img src=%s-R.png><BR>\n<img src=legend.gif><br>\n<BR>\n", Buffer1, Buffer1, HostName, Buffer1, Buffer1);					
					}
				}
			}

		fprintf(file, "</BODY></HTML>\n");
		fclose(file);
		}

	free(IPCData);
	}

void GraphIp(struct IPDataStore *DataStore, struct IPCData *IPCData, long int timestamp)
    {
    FILE *OutputFile;
    char outputfilename[50];
    gdImagePtr im, im2;
    int white;
    long int YMax;
	char CharIp[20];

    long int GraphBeginTime;

	// TODO: First determine if graph will be printed before creating image and drawing backround, etc

	HostIp2CharIp(DataStore->ip, CharIp);

    GraphBeginTime = timestamp - RANGE1;

    im = gdImageCreate(XWIDTH, YHEIGHT);
    white = gdImageColorAllocate(im, 255, 255, 255);
    //gdImageFill(im, 10, 10, white);

    im2 = gdImageCreate(XWIDTH, YHEIGHT);
    white = gdImageColorAllocate(im2, 255, 255, 255);
    //gdImageFill(im2, 10, 10, white);

    YMax = GraphData(im, im2, DataStore, GraphBeginTime, IPCData);
    if (YMax != 0)
        {
        // Finish the graph
        PrepareXAxis(im, timestamp);
        PrepareYAxis(im, YMax);

        PrepareXAxis(im2, timestamp);
        PrepareYAxis(im2, YMax);

        sprintf(outputfilename, "./htdocs/%s-S.png", CharIp);
        OutputFile = fopen(outputfilename, "w");    
        gdImagePng(im, OutputFile);
        fclose(OutputFile);

        sprintf(outputfilename, "./htdocs/%s-R.png", CharIp);
        OutputFile = fopen(outputfilename, "w");
        gdImagePng(im2, OutputFile);
        fclose(OutputFile);
        }
    else
        {
        // The graph isn't worth clutering up the web pages with
        sprintf(outputfilename, "./htdocs/%s-R.png", CharIp);
        unlink(outputfilename);
        sprintf(outputfilename, "./htdocs/%s-S.png", CharIp);
        unlink(outputfilename);
        }

	gdImageDestroy(im);
	gdImageDestroy(im2);
    }

// Returns YMax
long int GraphData(gdImagePtr im, gdImagePtr im2, struct IPDataStore *DataStore, long int timestamp, struct IPCData *IPCData)
    {
    long int YMax=0;
	
	struct DataStoreBlock *CurrentBlock;
    struct IPData *Data;

	// TODO: These should be a structure!!!!
	// TODO: This is an awfull lot of data to be allocated on the stack

    unsigned long total[XWIDTH];
    unsigned long icmp[XWIDTH];
    unsigned long udp[XWIDTH];
    unsigned long tcp[XWIDTH];
    unsigned long http[XWIDTH];
    unsigned long p2p[XWIDTH];
    int Count[XWIDTH];

    unsigned long total2[XWIDTH];
    unsigned long icmp2[XWIDTH];
    unsigned long udp2[XWIDTH];
    unsigned long tcp2[XWIDTH];
    unsigned long http2[XWIDTH];
    unsigned long p2p2[XWIDTH];

    size_t DataPoints;
    double x;
    int xint;
    int Counter;
    char Buffer[30];
    char Buffer2[50];
    
    int blue, red, yellow, purple, green, brown, black;
    int blue2, red2, yellow2, purple2, green2, brown2, black2;

	// Replaced by IPCData Structure
	//unsigned long int TotalSent = 0;
	//unsigned long int TotalReceived = 0;

	unsigned long int SentPeak = 0;
	unsigned long int ReceivedPeak = 0;

    yellow   = gdImageColorAllocate(im, 255, 255, 0);
    purple   = gdImageColorAllocate(im, 255, 0, 255);
    green    = gdImageColorAllocate(im, 0, 255, 0);
    blue     = gdImageColorAllocate(im, 0, 0, 255);
    brown    = gdImageColorAllocate(im, 128, 0, 0);
    red      = gdImageColorAllocate(im, 255, 0, 0);
    black 	 = gdImageColorAllocate(im, 0, 0, 0);
    
    yellow2  = gdImageColorAllocate(im2, 255, 255, 0);
    purple2   = gdImageColorAllocate(im2, 255, 0, 255);
    green2   = gdImageColorAllocate(im2, 0, 255, 0);
    blue2    = gdImageColorAllocate(im2, 0, 0, 255);
    brown2   = gdImageColorAllocate(im2, 128, 0, 0);
    red2     = gdImageColorAllocate(im2, 255, 0, 0);
    black2   = gdImageColorAllocate(im2, 0, 0, 0);

	CurrentBlock = DataStore->FirstBlock;
	Data = CurrentBlock->Data;
    DataPoints = CurrentBlock->NumEntries;

	memset(IPCData, 0, sizeof(struct IPCData));
	IPCData->IP = Data[0].ip;
	
    memset(Count, 0, sizeof(Count[0])*XWIDTH);

    memset(total, 0, sizeof(total[0])*XWIDTH);
    memset(icmp, 0, sizeof(total[0])*XWIDTH);
    memset(udp, 0, sizeof(total[0])*XWIDTH);
    memset(tcp, 0, sizeof(total[0])*XWIDTH);
    memset(http, 0, sizeof(total[0])*XWIDTH);
    memset(p2p, 0, sizeof(total[0])*XWIDTH);

    memset(total2, 0, sizeof(total[0])*XWIDTH);
    memset(icmp2, 0, sizeof(total[0])*XWIDTH);
    memset(udp2, 0, sizeof(total[0])*XWIDTH);
    memset(tcp2, 0, sizeof(total[0])*XWIDTH);
    memset(http2, 0, sizeof(total[0])*XWIDTH);
    memset(p2p2, 0, sizeof(total[0])*XWIDTH);

	// Change this to just run through all the datapoints we have stored in ram

	// Sum up the bytes/second
    while(DataPoints > 0)  // We have data to graph
        {
        for (Counter = 0; Counter < DataPoints; Counter++)  // Graph it all
            {
            x = (Data[Counter].timestamp-timestamp)*((XWIDTH-XOFFSET)/RANGE1)+XOFFSET;        
            xint = x;

            if (xint >= 0 && xint < XWIDTH)
                {
                Count[xint]++;
				
				if (Data[Counter].Send.total > SentPeak)
					SentPeak = Data[Counter].Send.total;
                total[xint] += Data[Counter].Send.total;
                icmp[xint] += Data[Counter].Send.icmp;
                udp[xint] += Data[Counter].Send.udp;
                tcp[xint] += Data[Counter].Send.tcp;
                http[xint] += Data[Counter].Send.http;
				p2p[xint] += Data[Counter].Send.p2p;

                if (Data[Counter].Receive.total > ReceivedPeak)
                	ReceivedPeak = Data[Counter].Receive.total;
                total2[xint] += Data[Counter].Receive.total;
                icmp2[xint] += Data[Counter].Receive.icmp;
                udp2[xint] += Data[Counter].Receive.udp;
                tcp2[xint] += Data[Counter].Receive.tcp;
                http2[xint] += Data[Counter].Receive.http;
				p2p2[xint] += Data[Counter].Receive.p2p;
                }
            }

		CurrentBlock = CurrentBlock->Next;
			
		if (CurrentBlock)
			{
         	Data = CurrentBlock->Data;
			DataPoints = CurrentBlock->NumEntries;
			}
		else
			DataPoints = 0;		
        }

	// Convert SentPeak and ReceivedPeak from bytes to bytes/second
	SentPeak /= INTERVAL; ReceivedPeak /= INTERVAL;

    // Preform the Average
    for(Counter=XOFFSET+1; Counter < XWIDTH; Counter++)
            {
            if (Count[Counter] > 0)
                {
            	IPCData->Total += total[Counter] + total2[Counter];
				IPCData->TotalSent += total[Counter];
 				IPCData->TotalReceived += total2[Counter];
				IPCData->TCP += tcp[Counter] + tcp2[Counter];
				IPCData->HTTP += http[Counter] + http2[Counter];
				IPCData->P2P += p2p[Counter] + p2p2[Counter];
				IPCData->UDP += udp[Counter] + udp2[Counter];
				IPCData->ICMP += icmp[Counter] + icmp2[Counter];

                // Preform the average
                total[Counter] /= (Count[Counter]*INTERVAL);
                tcp[Counter] /= (Count[Counter]*INTERVAL);
                http[Counter] /= (Count[Counter]*INTERVAL);
				p2p[Counter] /= (Count[Counter]*INTERVAL);
                udp[Counter] /= (Count[Counter]*INTERVAL);
                icmp[Counter] /= (Count[Counter]*INTERVAL);
								
                total2[Counter] /= (Count[Counter]*INTERVAL);
                tcp2[Counter] /= (Count[Counter]*INTERVAL);
                http2[Counter] /= (Count[Counter]*INTERVAL);
				p2p2[Counter] /= (Count[Counter]*INTERVAL);
                udp2[Counter] /= (Count[Counter]*INTERVAL);
                icmp2[Counter] /= (Count[Counter]*INTERVAL);


                if (total[Counter] > YMax)
                    YMax = total[Counter];
                
                if (total2[Counter] > YMax)
                    YMax = total2[Counter];
                }
            }

    YMax += YMax*0.05;    // Add an extra 5%
	
    if (IPCData->Total < config.graph_cutoff || !config.graph)
		{
		IPCData->Graph = FALSE;
        return(0);
		}
	else
        IPCData->Graph = TRUE;

    // Plot the points
    for(Counter=XOFFSET+1; Counter < XWIDTH; Counter++)    
            {
            if (Count[Counter] > 0)
                {
                // Convert the bytes/sec to y coords
                total[Counter] = (total[Counter]*(YHEIGHT-YOFFSET))/YMax;
                tcp[Counter] = (tcp[Counter]*(YHEIGHT-YOFFSET))/YMax;
                http[Counter] = (http[Counter]*(YHEIGHT-YOFFSET))/YMax;
                p2p[Counter] = (p2p[Counter]*(YHEIGHT-YOFFSET))/YMax;
                udp[Counter] = (udp[Counter]*(YHEIGHT-YOFFSET))/YMax;
                icmp[Counter] = (icmp[Counter]*(YHEIGHT-YOFFSET))/YMax;

                total2[Counter] = (total2[Counter]*(YHEIGHT-YOFFSET))/YMax;
                tcp2[Counter] = (tcp2[Counter]*(YHEIGHT-YOFFSET))/YMax;
                http2[Counter] = (http2[Counter]*(YHEIGHT-YOFFSET))/YMax;
				p2p2[Counter] = (p2p2[Counter]*(YHEIGHT-YOFFSET))/YMax;
                udp2[Counter] = (udp2[Counter]*(YHEIGHT-YOFFSET))/YMax;
                icmp2[Counter] = (icmp2[Counter]*(YHEIGHT-YOFFSET))/YMax;

                // Stack 'em up!
                // Total is stacked from the bottom
                // Icmp is on the bottom too
                // Udp is stacked on top of icmp
                udp[Counter] += icmp[Counter];
				udp2[Counter] += icmp2[Counter];
                // TCP and p2p are stacked on top of Udp
                tcp[Counter] += udp[Counter];
                tcp2[Counter] += udp2[Counter];
                p2p[Counter] += udp[Counter];
                p2p2[Counter] += udp2[Counter];
				// Http is stacked on top of p2p
                http[Counter] += p2p[Counter];
                http2[Counter] += p2p2[Counter];

                // Plot them!
				// Sent
                gdImageLine(im, Counter, (YHEIGHT-YOFFSET) - total[Counter], Counter, YHEIGHT-YOFFSET-1, yellow);
                gdImageLine(im, Counter, (YHEIGHT-YOFFSET) - icmp[Counter], Counter, YHEIGHT-YOFFSET-1, red);
                gdImageLine(im, Counter, (YHEIGHT-YOFFSET) - udp[Counter], Counter, (YHEIGHT-YOFFSET) - icmp[Counter] - 1, brown);
                gdImageLine(im, Counter, (YHEIGHT-YOFFSET) - tcp[Counter], Counter, (YHEIGHT-YOFFSET) - udp[Counter] - 1, green);
                gdImageLine(im, Counter, (YHEIGHT-YOFFSET) - p2p[Counter], Counter, (YHEIGHT-YOFFSET) - udp[Counter] - 1, purple);
                gdImageLine(im, Counter, (YHEIGHT-YOFFSET) - http[Counter], Counter, (YHEIGHT-YOFFSET) - p2p[Counter] - 1, blue);
				
				if (SentPeak < 1024/8)
					snprintf(Buffer2, 50, "Peak Send Rate: %.1f Bits/sec", (float)SentPeak*8);
				else if (SentPeak < (1024*1024)/8)
					snprintf(Buffer2, 50, "Peak Send Rate: %.1f KBits/sec", ((float)SentPeak*8.0)/1024.0);
				else snprintf(Buffer2, 50, "Peak Send Rate: %.1f MBits/sec", ((float)SentPeak*8.0)/(1024.0*1024.0));
				
				
				if (IPCData->TotalSent < 1024)
					snprintf(Buffer, 30, "Sent %.1f Bytes", (float)IPCData->TotalSent);					
				else if (IPCData->TotalSent < 1024*1024)
					snprintf(Buffer, 30, "Sent %.1f KBytes", (float)IPCData->TotalSent/1024.0);
				else snprintf(Buffer, 30, "Sent %.1f MBytes", (float)IPCData->TotalSent/(1024.0*1024.0));

				gdImageString(im, gdFontSmall, XOFFSET+5,  YHEIGHT-20, Buffer, black);
				gdImageString(im, gdFontSmall, XWIDTH/2+XOFFSET/2,  YHEIGHT-20, Buffer2, black);				

				
				// Receive
                gdImageLine(im2, Counter, (YHEIGHT-YOFFSET) - total2[Counter], Counter, YHEIGHT-YOFFSET-1, yellow2);
                gdImageLine(im2, Counter, (YHEIGHT-YOFFSET) - icmp2[Counter], Counter, YHEIGHT-YOFFSET-1, red2);
                gdImageLine(im2, Counter, (YHEIGHT-YOFFSET) - udp2[Counter], Counter, (YHEIGHT-YOFFSET) - icmp2[Counter] - 1, brown2);
                gdImageLine(im2, Counter, (YHEIGHT-YOFFSET) - tcp2[Counter], Counter, (YHEIGHT-YOFFSET) - udp2[Counter] - 1, green2);
                gdImageLine(im2, Counter, (YHEIGHT-YOFFSET) - p2p2[Counter], Counter, (YHEIGHT-YOFFSET) - udp2[Counter] - 1, purple2);
                gdImageLine(im2, Counter, (YHEIGHT-YOFFSET) - http2[Counter], Counter, (YHEIGHT-YOFFSET) - p2p2[Counter] - 1, blue2);

                if (ReceivedPeak < 1024/8)
                    snprintf(Buffer2, 50, "Peak Receive Rate: %.1f Bits/sec", (float)ReceivedPeak*8);
                else if (ReceivedPeak < (1024*1024)/8)
                    snprintf(Buffer2, 50, "Peak Receive Rate: %.1f KBits/sec", ((float)ReceivedPeak*8.0)/1024.0);
                else snprintf(Buffer2, 50, "Peak Receive Rate: %.1f MBits/sec", ((float)ReceivedPeak*8.0)/(1024.0*1024.0));


                if (IPCData->TotalReceived < 1024)
                    snprintf(Buffer, 30, "Received %.1f Bytes", (float)IPCData->TotalReceived);
                else if (IPCData->TotalReceived < 1024*1024)
                    snprintf(Buffer, 30, "Received %.1f KBytes", (float)IPCData->TotalReceived/1024.0);
                else snprintf(Buffer, 30, "Received %.1f MBytes", (float)IPCData->TotalReceived/(1024.0*1024.0));
                                                                                                              
                gdImageString(im2, gdFontSmall, XOFFSET+5,  YHEIGHT-20, Buffer, black2);                
                gdImageString(im2, gdFontSmall, XWIDTH/2+XOFFSET/2,  YHEIGHT-20, Buffer2, black2);
                }
            }
    return(YMax);
    }

void PrepareYAxis(gdImagePtr im, long int YMax)
    {
    char buffer[20];


    int black;
    float YTic = 0;
    double y;
    long int YStep;
    
    black = gdImageColorAllocate(im, 0, 0, 0);
    gdImageLine(im, XOFFSET, 0, XOFFSET, YHEIGHT, black);

    YLegend = ' ';
    Divisor = 1;
    if (YMax > 1024)
        {
        Divisor = 1024;    // Display in K
        YLegend = 'k';
        }
    if (YMax > 1048576)
        {
        Divisor = 1024; // Display in M
        YLegend = 'm';
        }
    if (YMax > 1073741824)
        {
        Divisor = 1073741824; // Display in G
        YLegend = 'g';
        }

    YStep = YMax/10;
    if (YStep < 1)
        YStep=1;
    YTic=YStep;

    while (YTic < (YMax - YMax/10))
        {
        y = (YHEIGHT-YOFFSET)-((YTic*(YHEIGHT-YOFFSET))/YMax);        

        gdImageLine(im, XOFFSET, y, XWIDTH, y, black);        
        snprintf(buffer, 20, "%4.0f %cbits/s", (8*YTic)/Divisor, YLegend);
        gdImageString(im, gdFontSmall, 3, y-7, buffer, black);        

        YTic += YStep;
        }
    } 

void PrepareXAxis(gdImagePtr im, long int timestamp)
    {
    char buffer[100];
    int black, red;
    long int sample_begin, sample_end;    
    struct tm *timestruct;
    long int MarkTime;
    double x;
    
    sample_begin=timestamp-RANGE1;
    sample_end=sample_begin+INTERVAL;

    black = gdImageColorAllocate(im, 0, 0, 0);
    red   = gdImageColorAllocate(im, 255, 0, 0);

    gdImageLine(im, 0, YHEIGHT-YOFFSET, XWIDTH, YHEIGHT-YOFFSET, black);

    // ********************************************************************
    // ****  Write the red day seperator bars
    // ********************************************************************

    timestruct = localtime(&sample_begin);
    timestruct->tm_sec = 0;
    timestruct->tm_min = 0;
    timestruct->tm_hour = 0;
    MarkTime = mktime(timestruct);
            
    x = (MarkTime-sample_begin)*( ((double)(XWIDTH-XOFFSET)) / RANGE1) + XOFFSET;
    if (x < XOFFSET)
        {
        MarkTime += (24*60*60);
        x = (MarkTime-sample_begin)*((XWIDTH-XOFFSET)/RANGE1) + XOFFSET;
        }

    while (x < (XWIDTH-10))
        {
        // Day Lines
        gdImageLine(im, x, 0, x, YHEIGHT-YOFFSET, red);
        gdImageLine(im, x+1, 0, x+1, YHEIGHT-YOFFSET, red);

        timestruct = localtime(&MarkTime);
        strftime(buffer, 100, "%a, %b %d", timestruct);
        gdImageString(im, gdFontSmall, x-30,  YHEIGHT-YOFFSET+10, buffer, black);        

        // Calculate Next x
        MarkTime += (24*60*60);
        x = (MarkTime-sample_begin)*((XWIDTH-XOFFSET)/RANGE1) + XOFFSET;
        }
    

    // ********************************************************************
    // ****  Write the tic marks
    // ********************************************************************

    timestruct = localtime(&sample_begin);
    timestruct->tm_sec = 0;
    timestruct->tm_min = 0;
    timestruct->tm_hour = 0;
    MarkTime = mktime(timestruct)+(6*60*60);

    x = (MarkTime-sample_begin)*((XWIDTH-XOFFSET)/RANGE1) + XOFFSET;
    while (x < (XWIDTH-10))
        {
        if (x > XOFFSET) {
            gdImageLine(im, x, YHEIGHT-YOFFSET-5, x, YHEIGHT-YOFFSET+5, black);
            gdImageLine(im, x+1, YHEIGHT-YOFFSET-5, x+1, YHEIGHT-YOFFSET+5, black);
            }
        MarkTime += (6*60*60);
        x = (MarkTime-sample_begin)*((XWIDTH-XOFFSET)/RANGE1) + XOFFSET;
        }

    timestruct = localtime(&sample_begin);
    timestruct->tm_sec = 0;
    timestruct->tm_min = 0;
    timestruct->tm_hour = 0;
    MarkTime = mktime(timestruct)+(60*60);

    x = (MarkTime-sample_begin)*((XWIDTH-XOFFSET)/RANGE1) + XOFFSET;
    while (x < (XWIDTH-10))
        {
        if (x > XOFFSET) {
            gdImageLine(im, x, YHEIGHT-YOFFSET, x, YHEIGHT-YOFFSET+5, black);
            gdImageLine(im, x+1, YHEIGHT-YOFFSET, x+1, YHEIGHT-YOFFSET+5, black);
            }
        MarkTime+=(60*60);
        x = (MarkTime-sample_begin)*((XWIDTH-XOFFSET)/RANGE1) + XOFFSET;
        }
    }


